package cn.com.qmth.examcloud.core.oe.task.service.impl;

import cn.com.qmth.examcloud.api.commons.enums.FaceApiProvider;
import cn.com.qmth.examcloud.commons.util.StringUtil;
import cn.com.qmth.examcloud.core.oe.student.dao.ExamCaptureQueueRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.ExamCaptureRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.entity.ExamCaptureEntity;
import cn.com.qmth.examcloud.core.oe.student.dao.entity.ExamCaptureQueueEntity;
import cn.com.qmth.examcloud.core.oe.task.service.ExamCaptureQueueService;
import cn.com.qmth.examcloud.core.oe.task.service.bean.FaceApiParam;
import cn.com.qmth.examcloud.starters.face.verify.common.CommonUtils;
import cn.com.qmth.examcloud.starters.face.verify.common.FaceVerifyException;
import cn.com.qmth.examcloud.starters.face.verify.model.FaceResult;
import cn.com.qmth.examcloud.starters.face.verify.model.param.FaceApiType;
import cn.com.qmth.examcloud.starters.face.verify.model.param.FaceParam;
import cn.com.qmth.examcloud.starters.face.verify.model.param.ImageBase64Parm;
import cn.com.qmth.examcloud.starters.face.verify.model.param.ImageParm;
import cn.com.qmth.examcloud.starters.face.verify.service.FaceVerifyService;
import cn.com.qmth.examcloud.support.Constants;
import cn.com.qmth.examcloud.support.cache.CacheHelper;
import cn.com.qmth.examcloud.support.cache.bean.OrgPropertyCacheBean;
import cn.com.qmth.examcloud.support.cache.bean.StudentCacheBean;
import cn.com.qmth.examcloud.support.fss.FssFactory;
import cn.com.qmth.examcloud.support.fss.FssHelper;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Example;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;

import java.util.Date;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * 处理考试抓拍队列
 */
@Service("examCaptureQueueService")
public class ExamCaptureQueueServiceImpl implements ExamCaptureQueueService {

    private static final Logger log = LoggerFactory.getLogger(ExamCaptureQueueServiceImpl.class);

    @Autowired
    private JdbcTemplate jdbcTemplate;

    @Autowired
    private ExamCaptureQueueRepo examCaptureQueueRepo;

    @Autowired
    private ExamCaptureRepo examCaptureRepo;

    @Autowired
    private FaceVerifyService faceVerifyService;

    @Override
    public void handlerExamCaptureQueuesByExamRecordDataId(Long examRecordDataId, List<ExamCaptureQueueEntity> queues
            , FaceApiParam param, AtomicInteger successCount) {
        if (CollectionUtils.isEmpty(queues)) {
            return;
        }

        // 获取当前学生的底照信息
        Long studentId = queues.get(0).getStudentId();
        StudentCacheBean studentCache = CacheHelper.getStudent(studentId);
        if (StringUtils.isBlank(studentCache.getPhotoPath())) {
            for (ExamCaptureQueueEntity queue : queues) {
                queue.setErrorNum(param.getMaxErrorNum());
                queue.setErrorMsg("学生底照不存在");
                queue.setUpdateTime(new Date());
            }
            examCaptureQueueRepo.saveAll(queues);
            log.warn("跳过人脸比对，学生底照不存在！studentId:{} examRecordDataId:{}", studentId, examRecordDataId);
            return;
        }

        // 获取学校“抓拍时人脸比对API”配置
        boolean useBaiduApi = param.isUseBaiduApi();
        OrgPropertyCacheBean faceApiConfig = CacheHelper.getOrgProperty(studentCache.getRootOrgId(), Constants.FACE_COMPARE_API_PROVIDER);
        if (faceApiConfig.getHasValue() && FaceApiProvider.FACEPP.name().equals(faceApiConfig.getValue())) {
            useBaiduApi = false;
        }

        // 获取学校“百度人脸比对通过阈值”配置
        Double baiduExpectFaceCompareScore = null;
        OrgPropertyCacheBean faceScoreConfig = CacheHelper.getOrgProperty(studentCache.getRootOrgId(), Constants.BAIDU_EXPECT_FACE_COMPARE_SCORE);
        if (faceScoreConfig.getHasValue() && StringUtils.isNotEmpty(faceScoreConfig.getValue())) {
            baiduExpectFaceCompareScore = Double.parseDouble(faceScoreConfig.getValue());
        }

        log.warn("studentId:{} examRecordDataId:{} queueSize:{} useBaiduApi:{} compare:{} liveness:{} scoreConfig:{}",
                studentId, examRecordDataId, queues.size(), useBaiduApi, param.isUseLocalBaiduApiForFaceCompare(),
                param.isUseLocalBaiduApiForFaceLiveness(), baiduExpectFaceCompareScore);

        // 将学生底照图片转成Base64格式
        ImageParm basePhoto;
        if (studentCache.getPhotoPath().contains("upyun.qmth.com.cn")) {
            // 兼容又拍云旧图片路径
            basePhoto = new ImageBase64Parm(CommonUtils.toBase64(studentCache.getPhotoPath()));
            log.warn("old data basePhotoUrl:{}", studentCache.getPhotoPath());
        } else {
            String basePhotoPath = FssHelper.fixFilePath(studentCache.getPhotoPath());
            byte[] basePhotoBytes = FssFactory.getInstance().readFile(basePhotoPath);
            basePhoto = new ImageBase64Parm(Base64.encodeBase64String(basePhotoBytes));
        }

        for (ExamCaptureQueueEntity queue : queues) {
            // 将抓拍照图片转成Base64格式
            ImageParm capturePhoto = this.checkAndReadFile(queue.getFileUrl(), examRecordDataId);
            if (capturePhoto == null) {
                // 删除错误无效数据
                examCaptureQueueRepo.deleteById(queue.getId());
                continue;
            }

            queue.setFaceCompareStartTime(System.currentTimeMillis());
            int curErrorNum = queue.getErrorNum() != null ? queue.getErrorNum() : 0;

            // 执行人脸比对
            boolean compareFinished = this.doFaceCompare(queue, basePhoto, capturePhoto, param, useBaiduApi, baiduExpectFaceCompareScore, curErrorNum);

            // 执行人脸真实性检测
            boolean verifyFinished = this.doFaceVerify(queue, capturePhoto, param, curErrorNum);

            if (compareFinished && verifyFinished) {
                // 保存最终结果并删除该队列记录
                this.saveExamCaptureAndDeleteQueue(queue);
                successCount.incrementAndGet();
            } else {
                examCaptureQueueRepo.save(queue);
            }

            capturePhoto = null;// 清理大对象内存
        }

        basePhoto = null;// 清理大对象内存
    }

    private ImageParm checkAndReadFile(String fileUrl, Long examRecordDataId) {
        String imgPath = FssHelper.fixFilePath(fileUrl);
        if (StringUtils.isBlank(imgPath)) {
            log.warn("照片地址有误！examRecordDataId:{} imgUrl:{}", examRecordDataId, fileUrl);
            return null;
        }

        byte[] imgBytes;
        try {
            imgBytes = FssFactory.getInstance().readFile(imgPath);
        } catch (Exception e) {
            boolean exist = FssFactory.getInstance().existFile(imgPath);
            if (!exist) {
                log.warn("照片文件不存在！examRecordDataId:{} imgUrl:{}", examRecordDataId, fileUrl);
                return null;
            }
            throw e;
        }

        if (imgBytes == null || imgBytes.length == 0) {
            log.warn("照片文件有误！examRecordDataId:{} imgUrl:{}", examRecordDataId, fileUrl);
            return null;
        }

        return new ImageBase64Parm(Base64.encodeBase64String(imgBytes));
    }

    private boolean doFaceCompare(ExamCaptureQueueEntity queue, ImageParm basePhoto, ImageParm capturePhoto,
                                  FaceApiParam param, boolean useBaiduApi, Double baiduExpectFaceCompareScore,
                                  int curErrorNum) {
        if (StringUtils.isNotBlank(queue.getFaceCompareResult())) {
            // 已执行过“人脸比对”，则跳过
            return true;
        }

        try {
            // 人脸比对开始处理
            FaceResult faceCompareResult;
            if (useBaiduApi) {
                faceCompareResult = faceVerifyService.faceCompare(FaceParam.builder()
                        .images(basePhoto, capturePhoto)
                        .expectFaceCompareScore(baiduExpectFaceCompareScore)
                        .apiType(param.isUseLocalBaiduApiForFaceCompare() ? FaceApiType.PRIVATE_BAIDU_API : FaceApiType.BAIDU_API));
            } else {
                faceCompareResult = faceVerifyService.faceCompare(FaceParam.builder()
                        .images(basePhoto, capturePhoto)
                        .apiType(FaceApiType.FACE_PLUS));
            }

            if (faceCompareResult.isApiNeedRetry()) {
                throw new FaceVerifyException(faceCompareResult.getError());
            }

            queue.setIsPass(faceCompareResult.isFacePass());
            queue.setFaceCompareResult(faceCompareResult.getJsonResult());
            queue.setIsStranger(faceCompareResult.getFaceNum() > 1);
            return true;
        } catch (Exception e) {
            queue.setErrorNum(curErrorNum + 1);
            queue.setErrorMsg(StringUtil.substring("人脸比对异常！" + e.getMessage(), 200));

            log.error("人脸比对异常！examRecordDataId:{} queueId:{} errorNum:{} error:{}",
                    queue.getExamRecordDataId(), queue.getId(), queue.getErrorNum(), e.getMessage());
            return false;
        }
    }

    private boolean doFaceVerify(ExamCaptureQueueEntity queue, ImageParm capturePhoto, FaceApiParam param, int curErrorNum) {
        if (StringUtils.isNotBlank(queue.getFacelivenessResult())) {
            // 已执行过“真实性检测”，则跳过
            return true;
        }

        try {
            // 人脸真实性检测开始
            FaceResult faceLivenessResult = faceVerifyService.faceVerify(FaceParam.builder().images(capturePhoto)
                    .apiType(param.isUseLocalBaiduApiForFaceLiveness() ? FaceApiType.PRIVATE_BAIDU_API : FaceApiType.BAIDU_API));

            if (faceLivenessResult.isApiNeedRetry()) {
                throw new FaceVerifyException(faceLivenessResult.getError());
            }

            queue.setFacelivenessResult(faceLivenessResult.getJsonResult());
            return true;
        } catch (Exception e) {
            queue.setErrorNum(curErrorNum + 1);
            queue.setErrorMsg(StringUtil.substring("人脸真实性检测异常！" + e.getMessage(), 200));

            log.error("人脸真实性检测异常！examRecordDataId:{} queueId:{} errorNum:{} error:{}",
                    queue.getExamRecordDataId(), queue.getId(), queue.getErrorNum(), e.getMessage());
            return false;
        }
    }

    private void saveExamCaptureAndDeleteQueue(ExamCaptureQueueEntity queue) {
        ExamCaptureEntity query = new ExamCaptureEntity();
        query.setExamRecordDataId(queue.getExamRecordDataId());
        query.setFileName(queue.getFileName());
        if (examCaptureRepo.exists(Example.of(query))) {
            // 照片处理结果中若已存在则以已有的数据为准，删除该队列记录
            examCaptureQueueRepo.deleteById(queue.getId());
            log.info("RemoveQueue examRecordDataId:{} fileName:{}", queue.getExamRecordDataId(), queue.getFileName());
            return;
        }

        ExamCaptureEntity result = new ExamCaptureEntity();
        result.setExamRecordDataId(queue.getExamRecordDataId());
        result.setFileName(queue.getFileName());
        result.setFileUrl(queue.getFileUrl());
        result.setIsPass(queue.getIsPass());
        result.setIsStranger(queue.getIsStranger());
        result.setFaceCompareResult(queue.getFaceCompareResult());
        result.setFacelivenessResult(queue.getFacelivenessResult());
        result.setExtMsg(queue.getExtMsg());

        long currentTimeMillis = System.currentTimeMillis();
        // 从创建队列到处理完毕的时间
        result.setProcessTime(currentTimeMillis - queue.getCreationTime().getTime());
        // 从开始处理到处理完毕的时间
        result.setUsedTime(currentTimeMillis - queue.getFaceCompareStartTime());

        result.setHasVirtualCamera(queue.getHasVirtualCamera() != null ? queue.getHasVirtualCamera() : false);
        if (StringUtils.length(queue.getCameraInfos()) >= Constants.VM_CAMERA_SIZE_LIMIT) {
            log.warn("虚拟摄像头信息超长！examRecordDataId:{} queueId:{}", queue.getExamRecordDataId(), queue.getId());
            result.setCameraInfos(Constants.VM_CAMERA_WARN);
        } else {
            result.setCameraInfos(queue.getCameraInfos());
        }

        examCaptureRepo.save(result);

        // 删除该队列记录
        examCaptureQueueRepo.deleteById(queue.getId());
        log.info("RemoveQueue examRecordDataId:{} fileName:{}", queue.getExamRecordDataId(), queue.getFileName());
    }

    /**
     * 按考试记录ID分片获取待处理的抓拍照片记录
     *
     * @param shardTotal      分片总数
     * @param shardIndex      当前分片索引
     * @param batchSize       批量条数
     * @param maxErrorNum     最大重试错误次数
     * @param examHandInFirst 是否优先处理“已交卷”的考试记录
     * @return 考试记录ID列表
     */
    @Override
    public List<Long> findQueuesGroupByExamRecordDataId(int shardTotal, int shardIndex, int batchSize, int maxErrorNum, boolean examHandInFirst) {
        StringBuilder sql = new StringBuilder();
        sql.append(" select q.exam_record_data_id");
        sql.append(" from ec_oet_exam_capture_queue q");
        if (examHandInFirst) {
            sql.append(" inner join ec_oes_exam_record_data d on d.id = q.exam_record_data_id");
        }
        sql.append(" where mod(q.exam_record_data_id,").append(shardTotal).append(") = ").append(shardIndex);
        sql.append(" and q.error_num < ").append(maxErrorNum);
        if (examHandInFirst) {
            sql.append(" and d.sync_status = 'UNSYNC'");
            sql.append(" and d.exam_record_status in('EXAM_HAND_IN','EXAM_AUTO_HAND_IN')");
        }
        sql.append(" group by q.exam_record_data_id");
        sql.append(" order by q.exam_record_data_id asc");
        sql.append(" limit ").append(batchSize);

        if (log.isDebugEnabled()) {
            log.debug(sql.toString());
        }

        return jdbcTemplate.queryForList(sql.toString(), Long.class);
    }

}
