package cn.com.qmth.examcloud.core.oe.task.service.job;

import cn.com.qmth.examcloud.core.oe.student.dao.ExamRecordDataRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.entity.ExamRecordDataEntity;
import cn.com.qmth.examcloud.core.oe.task.service.ExamRecordDataService;
import cn.com.qmth.examcloud.support.CacheConstants;
import cn.com.qmth.examcloud.support.enums.ExamRecordStatus;
import cn.com.qmth.examcloud.support.examing.ExamRecordData;
import cn.com.qmth.examcloud.web.exception.SequenceLockException;
import cn.com.qmth.examcloud.web.helpers.SequenceLockHelper;
import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.List;

/**
 * 处理交卷后考试数据任务
 */
@Component
public class AfterHandInExamJobHandler {

    private static final Logger log = LoggerFactory.getLogger(AfterHandInExamJobHandler.class);

    @Autowired
    private ExamRecordDataRepo examRecordDataRepo;

    @Autowired
    private ExamRecordDataService examRecordDataService;

    public void run(int shardTotal, int shardIndex, String jobParam) throws Exception {
        int pageNo = 0, pageSize = 200, sumTodoSize = 0;
        Long startId = 0L;

        while (true) {
            List<ExamRecordDataEntity> todoList = examRecordDataService.getExamRecordDataUnSyncList(
                    shardTotal, shardIndex, pageSize, startId, ExamRecordStatus.EXAM_HAND_IN, ExamRecordStatus.EXAM_AUTO_HAND_IN);

            pageNo++;
            sumTodoSize += todoList.size();
            if (CollectionUtils.isEmpty(todoList)) {
                log.warn("   分片任务_交卷后_{}_{} 本次处理考试记录数:{}", shardTotal, shardIndex, sumTodoSize);
                break;
            }

            log.warn("   分片任务_交卷后_{}_{} pageNo:{} 当前条数:{} 累计条数:{}", shardTotal, shardIndex, pageNo, todoList.size(), sumTodoSize);
            int curTodoIndex = 0;
            for (ExamRecordDataEntity data : todoList) {
                final String lockKey = CacheConstants.LOCK_EXAM_CONTROL + data.getStudentId();
                try {
                    SequenceLockHelper.getLockSimple(lockKey);

                    // 处理业务
                    curTodoIndex++;
                    this.handler(data.getId(), todoList.size(), curTodoIndex);
                } catch (Exception e) {
                    if (e instanceof InterruptedException) {
                        // 若线程终止，则抛出交由任务调度中心处理
                        log.warn("当前任务线程被终止！examRecordDataId:{}, error:{}", data.getId(), e.getMessage());
                        throw e;
                    } else if (e instanceof SequenceLockException) {
                        // 若锁问题，下次会继续执行
                        log.warn("当前任务获取锁失败！examRecordDataId:{}, redisKey:{}", data.getId(), lockKey);
                    } else {
                        // 若异常，下次会继续执行（需要排查原因）
                        log.error("当前任务处理失败！examRecordDataId:{}, error:{}", data.getId(), e.getMessage());
                    }
                } finally {
                    SequenceLockHelper.releaseLockSimple(lockKey);
                }
            }

            startId = todoList.get(todoList.size() - 1).getId();
            todoList.clear();
        }
    }

    /**
     * 仅处理已交卷的考试记录
     */
    private void handler(Long examRecordDataId, int curTodoSize, int curTodoIndex) throws Exception {
        ExamRecordData examRecordDataCache = examRecordDataService.getExamRecordDataCache(examRecordDataId);
        if (examRecordDataCache == null) {
            log.warn("redisKey {}{} value not exist, Set examRecordStatus=EXAM_ERROR", CacheConstants.CACHE_OE_RECORD, examRecordDataId);
            examRecordDataRepo.updateExamRecordStatusById(ExamRecordStatus.EXAM_ERROR, new Date(), examRecordDataId);
            return;
        }

        if (ExamRecordStatus.EXAM_HAND_IN != examRecordDataCache.getExamRecordStatus()
                && ExamRecordStatus.EXAM_AUTO_HAND_IN != examRecordDataCache.getExamRecordStatus()) {
            log.warn("Mysql&Redis examRecordStatus value not equal, should EXAM_HAND_IN or EXAM_AUTO_HAND_IN but {}, examRecordDataId:{}",
                    examRecordDataCache.getExamRecordStatus(), examRecordDataId);
            // 状态出现不一致情况时，则按Redis中的值为准，下次会按实际状态执行
            examRecordDataRepo.updateExamRecordStatusById(examRecordDataCache.getExamRecordStatus(), new Date(), examRecordDataId);
            return;
        }

        // 交卷时间（手动交卷取endTime，自动交卷取cleanTime）
        Date handInTime = examRecordDataCache.getEndTime() != null ? examRecordDataCache.getEndTime() : examRecordDataCache.getCleanTime();
        long diffSecond = (System.currentTimeMillis() - handInTime.getTime()) / 1000L;
        String warnMsg = (diffSecond / 3600L) >= 12L ? "【警告】交卷处理超过12小时！！！" : "";
        log.warn("【{}】交卷距离现在已:{}, examRecordDataId:{}, studentId:{}, curTodoSize:{}, curTodoIndex:{} {}",
                examRecordDataCache.getHandInExamType(), this.diff(diffSecond), examRecordDataId,
                examRecordDataCache.getStudentId(), curTodoSize, curTodoIndex, warnMsg);

        try {
            // 处理人脸、活体、违纪、算分等
            examRecordDataService.processAfterHandInExam(examRecordDataCache);
        } catch (Exception e) {
            log.error("processAfterHandInExam fail, examRecordDataId:{}, error:{}", examRecordDataCache.getId(), e.getMessage());
            throw e;
        }
    }

    /**
     * 将相差多少秒 格式化为：xx小时xx分xx秒
     */
    private String diff(long diffSecond) {
        long hour = diffSecond / 3600;
        long minute = diffSecond / 60 - hour * 60;
        long second = diffSecond - hour * 3600 - minute * 60;

        StringBuilder result = new StringBuilder();
        if (hour > 0) {
            result.append(hour).append("小时");
            result.append(minute).append("分");
            result.append(second).append("秒");
            return result.toString();
        }
        if (minute > 0) {
            result.append(minute).append("分");
            result.append(second).append("秒");
            return result.toString();
        }
        result.append(second).append("秒");
        return result.toString();
    }

}
