package cn.com.qmth.examcloud.core.oe.task.service.job;

import cn.com.qmth.examcloud.core.oe.admin.api.SyncExamDataCloudService;
import cn.com.qmth.examcloud.core.oe.admin.api.bean.ExamCaptureBean;
import cn.com.qmth.examcloud.core.oe.admin.api.bean.ExamSyncCaptureBean;
import cn.com.qmth.examcloud.core.oe.admin.api.request.SyncExamDataReq;
import cn.com.qmth.examcloud.core.oe.student.base.OeConstants;
import cn.com.qmth.examcloud.core.oe.student.dao.ExamRecordDataRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.entity.ExamRecordDataEntity;
import cn.com.qmth.examcloud.core.oe.task.service.CommonService;
import cn.com.qmth.examcloud.core.oe.task.service.ExamBossService;
import cn.com.qmth.examcloud.core.oe.task.service.ExamRecordDataService;
import cn.com.qmth.examcloud.core.oe.task.service.ExamingSessionService;
import cn.com.qmth.examcloud.support.CacheConstants;
import cn.com.qmth.examcloud.support.enums.ExamRecordStatus;
import cn.com.qmth.examcloud.support.enums.SyncStatus;
import cn.com.qmth.examcloud.support.examing.ExamBoss;
import cn.com.qmth.examcloud.support.examing.ExamRecordData;
import cn.com.qmth.examcloud.support.examing.ExamingSession;
import cn.com.qmth.examcloud.support.helper.FaceBiopsyHelper;
import cn.com.qmth.examcloud.support.redis.RedisKeyHelper;
import cn.com.qmth.examcloud.web.exception.SequenceLockException;
import cn.com.qmth.examcloud.web.helpers.SequenceLockHelper;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * 同步考试记录数据任务
 */
@Component
public class SyncExamRecordDataJobHandler {

    private static final Logger log = LoggerFactory.getLogger(SyncExamRecordDataJobHandler.class);

    @Autowired
    private ExamRecordDataRepo examRecordDataRepo;

    @Autowired
    private ExamRecordDataService examRecordDataService;

    @Autowired
    private ExamBossService examBossService;

    @Autowired
    private SyncExamDataCloudService syncExamDataCloudService;

    @Autowired
    private ExamingSessionService examingSessionService;

    @Autowired
    private CommonService commonService;

    @Autowired
    private RedisClient redisClient;

    public void run(int shardTotal, int shardIndex, String jobParam) throws Exception {
        int pageNo = 0, pageSize = 200, sumTodoSize = 0;
        Long startId = 0L;

        while (true) {
            List<ExamRecordDataEntity> todoList = examRecordDataService.getExamRecordDataUnSyncList(
                    shardTotal, shardIndex, pageSize, startId, ExamRecordStatus.EXAM_END, ExamRecordStatus.EXAM_OVERDUE);

            pageNo++;
            sumTodoSize += todoList.size();
            if (CollectionUtils.isEmpty(todoList)) {
                log.warn("分片任务_同步归档_{}_{} 本次处理考试记录数:{}", shardTotal, shardIndex, sumTodoSize);
                break;
            }

            log.warn("分片任务_同步归档_{}_{} pageNo:{} 当前条数:{} 累计条数:{}", shardTotal, shardIndex, pageNo, todoList.size(), sumTodoSize);
            for (ExamRecordDataEntity data : todoList) {
                final String lockKey = CacheConstants.LOCK_EXAM_CONTROL + data.getStudentId();
                try {
                    SequenceLockHelper.getLockSimple(lockKey);

                    // 处理业务
                    this.handler(data.getId());
                } catch (Exception e) {
                    if (e instanceof InterruptedException) {
                        // 若线程终止，则抛出交由任务调度中心处理
                        log.warn("当前任务线程被终止！examRecordDataId:{}, error:{}", data.getId(), e.getMessage());
                        throw e;
                    } else if (e instanceof SequenceLockException) {
                        // 若锁问题，下次会继续执行
                        log.warn("当前任务获取锁失败！examRecordDataId:{}, redisKey:{}", data.getId(), lockKey);
                    } else {
                        // 若异常，下次会继续执行（需要排查原因）
                        log.error("当前任务处理失败！examRecordDataId:{}, error:{}", data.getId(), e.getMessage());
                    }
                } finally {
                    SequenceLockHelper.releaseLockSimple(lockKey);
                }
            }

            startId = todoList.get(todoList.size() - 1).getId();
            todoList.clear();
        }
    }

    /**
     * 仅处理已结束的考试记录
     */
    private void handler(Long examRecordDataId) throws Exception {
        ExamRecordData examRecordDataCache = examRecordDataService.getExamRecordDataCache(examRecordDataId);
        if (examRecordDataCache == null) {
            log.warn("redisKey {}{} value not exist, Set examRecordStatus=EXAM_ERROR", CacheConstants.CACHE_OE_RECORD, examRecordDataId);
            examRecordDataRepo.updateExamRecordStatusById(ExamRecordStatus.EXAM_ERROR, new Date(), examRecordDataId);
            return;
        }

        if (ExamRecordStatus.EXAM_END != examRecordDataCache.getExamRecordStatus()
                && ExamRecordStatus.EXAM_OVERDUE != examRecordDataCache.getExamRecordStatus()) {
            log.warn("Mysql&Redis examRecordStatus value not equal, should EXAM_END or EXAM_OVERDUE but {}, examRecordDataId:{}",
                    examRecordDataCache.getExamRecordStatus(), examRecordDataId);
            // 状态出现不一致情况时，则按Redis中的值为准，下次会按实际状态执行
            examRecordDataRepo.updateExamRecordStatusById(examRecordDataCache.getExamRecordStatus(), new Date(), examRecordDataId);
            return;
        }

        if (SyncStatus.SYNCED == examRecordDataCache.getSyncStatus()) {
            log.warn("syncExamData already synced, examRecordDataId:{}, studentId:{}", examRecordDataId, examRecordDataCache.getStudentId());
            this.afterSyncExamData(examRecordDataCache);
            return;
        }

        log.info("syncExamData start, examRecordDataId:{}, studentId:{}", examRecordDataId, examRecordDataCache.getStudentId());

        SyncExamDataReq syncReq = new SyncExamDataReq();
        // 获取试卷结构（Redis中）
        syncReq.setExamRecordPaperStruct(commonService.getExamRecordPaperStruct(examRecordDataId));
        // 获取考试作答记录（Redis中）
        syncReq.setExamRecordQuestions(commonService.getExamRecordQuestions(examRecordDataId));
        // 获取考试过程记录（Mysql中）
        syncReq.setExamProcessRecords(commonService.getExamProcessRecords(examRecordDataId));
        // 获取断点续考记录（Mysql中）
        syncReq.setExamContinuedRecords(commonService.getExamContinuedRecords(examRecordDataId));

        Long rootOrgId = examRecordDataCache.getRootOrgId();
        Long examId = examRecordDataCache.getExamId();
        Long studentId = examRecordDataCache.getStudentId();

        // 是否开启人脸验证
        if (FaceBiopsyHelper.isFaceEnable(rootOrgId, examId, studentId)) {
            // 获取同步抓拍照片处理结果（Mysql中）
            ExamSyncCaptureBean syncCapture = commonService.getExamSyncCapture(examRecordDataId);
            syncReq.setExamSyncCapture(syncCapture);

            // 获取间隔抓拍照片处理记录（Mysql中）
            List<ExamCaptureBean> captures = commonService.getExamCaptures(examRecordDataId);
            syncReq.setExamCaptures(captures);

            // 是否开启考中人脸活体检测
            if (FaceBiopsyHelper.isFaceVerify(rootOrgId, examId, studentId)) {
                // S1 FACE_ID FaceID活体检测方案(即旧活体检测方案)
                syncReq.setExamFaceLivenessVerifies(commonService.getExamFaceLivenessVerifies(examRecordDataId));

                // S2 FACE_MOTION 自研活体检测方案【已废弃】
                syncReq.setFaceBiopsy(commonService.getFaceBiopsy(examRecordDataId));

                // S3 FACE_CLIENT C端活体检测方案
                syncReq.setFaceLiveVerifyRecords(commonService.getFaceLiveVerifyRecords(examRecordDataId));
            }
        }

        // 考试记录信息
        syncReq.setExamRecordData(commonService.copyExamRecordDataFrom(examRecordDataCache));

        long startTime = System.currentTimeMillis();
        syncExamDataCloudService.syncExamData(syncReq);
        long cost = System.currentTimeMillis() - startTime;
        log.warn("syncExamData ok, examRecordDataId:{}, studentId:{}, examType:{}, cost:{}ms",
                examRecordDataId, examRecordDataCache.getStudentId(), examRecordDataCache.getExamType(), cost);

        // 同步成功后，更新同步状态
        examRecordDataCache.setSyncStatus(SyncStatus.SYNCED);
        examRecordDataService.saveExamRecordDataCache(examRecordDataId, examRecordDataCache);

        this.afterSyncExamData(examRecordDataCache);
    }

    private void afterSyncExamData(ExamRecordData examRecordDataCache) {
        // 更新数据库的同步状态
        examRecordDataRepo.updateSyncStatusById(SyncStatus.SYNCED, examRecordDataCache.getId());

        // 考生已考完结次数累加
        ExamBoss examBoss = examBossService.getExamBoss(examRecordDataCache.getExamStudentId());
        if (null != examBoss) {
            examBoss.setEndCount(examBoss.getEndCount() + 1);
            examBossService.saveExamBoss(examRecordDataCache.getExamStudentId(), examBoss);
            log.info("ExamBoss startCount:{}, endCount:{}, examRecordDataId:{}, studentId:{}", examBoss.getStartCount(),
                    examBoss.getEndCount(), examRecordDataCache.getId(), examRecordDataCache.getStudentId());
        }

        this.clearUnusefulCacheData(examRecordDataCache);
    }

    /**
     * 清理不再使用的缓存数据
     */
    private void clearUnusefulCacheData(ExamRecordData examRecordDataCache) {
        if (SyncStatus.SYNCED != examRecordDataCache.getSyncStatus()) {
            return;
        }

        Long studentId = examRecordDataCache.getStudentId();
        Long examStudentId = examRecordDataCache.getExamStudentId();
        Long examRecordDataId = examRecordDataCache.getId();

        // 清除当前考生的考试次数相关缓存
        final String examBossKey = RedisKeyHelper.getBuilder().examBossKey(examStudentId);
        ExamBoss examBoss = redisClient.get(examBossKey, ExamBoss.class);
        if (examBoss != null) {
            if (examBoss.getStartCount() == examBoss.getEndCount()) {
                // 如果开考次数 等于 考试完结次数，则删除该缓存
                redisClient.delete(examBossKey);
            } else {
                // 否则仅移除当前的考试记录ID
                List<Long> examRecordDataIds = examBoss.getExamRecordDataIds();
                if (CollectionUtils.isNotEmpty(examRecordDataIds)) {
                    examRecordDataIds.removeIf(e -> examRecordDataId.equals(e));
                    examBoss.setExamRecordDataIds(examRecordDataIds);
                    redisClient.set(examBossKey, examBoss, OeConstants.TIME_OUT_7_DAY);
                }
            }
        }

        // 清除当前“考试活动时间”缓存
        redisClient.delete(RedisKeyHelper.getBuilder().examingActiveTimeKey(examRecordDataId));

        // 清除心跳缓存
        redisClient.delete(RedisKeyHelper.getBuilder().examingHeartbeatKey(examRecordDataId));

        // 清除当前学生的考试会话缓存（交卷后，正好又重新开考“预处理阶段”时examingSession的examRecordDataId会为空！）
        ExamingSession examingSession = examingSessionService.getExamingSession(studentId);
        if (examingSession != null && examRecordDataId.equals(examingSession.getExamRecordDataId())) {
            examingSessionService.deleteExamingSession(studentId);
        }

        // 清除试题作答记录、作答文件记录缓存
        Integer questionCount = examRecordDataCache.getQuestionCount();
        for (int n = 1; n <= questionCount; n++) {
            redisClient.delete(RedisKeyHelper.getBuilder().studentAnswerKey(examRecordDataId, n));
            redisClient.delete(RedisKeyHelper.getBuilder().studentFileAnswerKey(examRecordDataId, n));
        }

        // 清除“网考试卷结构”缓存
        redisClient.delete(RedisKeyHelper.getBuilder().studentPaperKey(examRecordDataId));

        // 清除“违纪非法请求数据”缓存
        redisClient.delete(CacheConstants.CACHE_OE_DISCIPLINE_ILLEGAL_DATA + examRecordDataId);

        // 清除“违纪非法考生端应用”缓存
        redisClient.delete(CacheConstants.CACHE_OE_DISCIPLINE_ILLEGAL_CLIENT + examRecordDataId);

        // 最后设置过期“当前考试记录”缓存
        final String examRecordDataKey = RedisKeyHelper.getBuilder().examRecordDataKey(examRecordDataId);
        redisClient.expire(examRecordDataKey, OeConstants.TIME_OUT_5_MINUTE, TimeUnit.SECONDS);

        log.warn("clearUnusefulCacheData finished, examRecordDataId:{}, studentId:{}", examRecordDataId, studentId);
    }

}
